
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: pants_plugins/sendwave/pants_docker:pants_docker

from setuptools import setup

setup(**{
    'author': 'Nathan Rosenbloom, Jean Cochrane',
    'author_email': 'engineering@sendwave.com',
    'description': 'Pants Plugin to automatically generate docker images from pants targets',
    'install_requires': (
        'pantsbuild.pants.testutil<2.14,>=2.13.0a0',
        'pantsbuild.pants<2.14,>=2.13.0a0',
    ),
    'name': 'sendwave-pants-docker',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'sendwave.pants_docker',
    ),
    'python_requires': '==3.9.*',
    'url': 'https://github.com/waveremit/pants-docker',
    'version': '1.1.0',
})
